/* ceelog(1) command-line utility.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Author: Miloslav Trmač <mitr@redhat.com> */
#include <config.h>

#include <getopt.h>
#include <locale.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <libintl.h>

#include <ceelog/ceelog.h>
#include <ceelog/event.h>
#include <ceelog/expression.h>
#include <ceelog/source.h>

/* Only print the message text, not full event data */
static bool conf_message_only; /* = false; */

/* do not stop when EOF is reached */
static bool conf_follow; /* = false; */

/* Expression to search for */
struct ceelog_compiled_expr *conf_expr;

/* Print an error message FMT, ... (without trailing '\n') to stderr, and
   exit (EXIT_FAILURE). */
G_GNUC_NORETURN
static void
error_exit (const char *fmt, ...)
{
  va_list ap;

  fprintf (stderr, _("%s: "), g_get_prgname ());
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (EXIT_FAILURE);
}

/* Output --help text */
static void
help (void)
{
  printf (_("Usage: ceelog [OPTION]... [FILTER]\n"
	    "Search for log events.\n"
	    "\n"
	    "  -f, --follow           do not stop when end-of-file is reached\n"
	    "  -h, --help             print this help\n"
	    "  -m, --message          only print a readable message without "
	    "most metadata\n"
	    "  -V, --version          print version information\n"));
  printf (_("\n"
	    "Report bugs to %s.\n"), PACKAGE_BUGREPORT);
}

/* Parse options in ARGC, ARGV.  Exit on error or --help, --version. */
static void
parse_argv (int argc, char *argv[])
{
  static const struct option options[] =
    {
      { "follow", no_argument, NULL, 'f'},
      { "help", no_argument, NULL, 'h' },
      { "message", no_argument, NULL, 'm' },
      { "version", no_argument, NULL, 'V' },
      { NULL, 0, NULL, 0 }
    };

  GError *error;

  for (;;)
    {
      int opt, idx;

      opt = getopt_long (argc, argv, "Vfhm", options, &idx);
      switch (opt)
	{
	case -1:
	  goto options_done;

	case '?':
	  exit (EXIT_FAILURE);

	case 'V':
	  puts (PACKAGE_NAME " " PACKAGE_VERSION);
	  puts (_("Copyright (C) 2012 Red Hat, Inc. All rights reserved.\n"
		  "This software is distributed under the GPL v.2.\n"
		  "\n"
		  "This program is provided with NO WARRANTY, to the extent "
		  "permitted by law."));
	  exit (EXIT_SUCCESS);

	case 'h':
	  help ();
	  exit (EXIT_SUCCESS);

	case 'm':
	  conf_message_only = true;
	  break;

	case 'f':
	  conf_follow = true;
	  break;

	default:
	  g_assert_not_reached ();
	}
    }
 options_done:
  if (argc != optind + 1)
    error_exit (_("Filter expression expected"));
  error = NULL;
  conf_expr = ceelog_parse_and_compile_expr (argv[optind], &error);
  if (conf_expr == NULL)
    error_exit (_("Error compiling %s: %s"), argv[optind], error->message);
}

/* Output matching EVT. */
static void
output_event (struct ceelog_event *evt)
{
  if (conf_message_only)
    {
      char buf[BUFSIZ], *tag, *msg;
      time_t ts;

      ts = ceelog_event_get_timestamp (evt);
      if (ts != -1)
	{
	  struct tm *tm;

	  tm = localtime (&ts);
	  /* Note the trailing space, we won't need a separator below. */
	  if (tm == NULL || strftime (buf, sizeof (buf), "%c ", tm) == 0)
	    ts = -1;
	}
      if (ts == -1)
	*buf = '\0';

      tag = ceelog_event_get_tag (evt);
      msg = ceelog_event_get_message (evt);
      printf ("%s%s%s%s\n", buf, tag != NULL ? tag : "",
	      tag != NULL ? ": " : "", msg);
      g_free (msg);
      g_free (tag);
    }
  else
    {
      char *text;

      text = ceelog_event_to_string (evt);
      printf ("%s\n", text);
      g_free (text);
    }
}

int
main (int argc, char *argv[])
{
  GError *error;
  struct ceelog_source *src;
  struct ceelog_event *evt;

  setlocale (LC_ALL, "");
  textdomain (PACKAGE_NAME);
  bindtextdomain (PACKAGE_NAME, LOCALEDIR);
  ceelog_init ();

  /* Eventually this should do much more. */
  g_set_prgname ("ceelog");
  parse_argv (argc, argv);

  error = NULL;
  src = ceelog_source_new_default (&error);
  if (src == NULL)
    error_exit (_("Error opening event source: %s"), error->message);

  if (conf_follow)
    {
      while ((evt = ceelog_source_read_event (src, &error)) != NULL)
	{
	  if (ceelog_eval_expr (conf_expr, evt))
	    output_event (evt);
	  ceelog_event_free (evt);
	}
    }
  else
    {
      while ((evt = ceelog_source_read_event_nonblocking (src, &error)) != NULL)
	{
	  if (ceelog_eval_expr (conf_expr, evt))
	    output_event (evt);
	  ceelog_event_free (evt);
	}
    }

  if (error != NULL)
    {
      if (!g_error_matches (error, CEELOG_ERROR, CEELOG_ERROR_WOULD_BLOCK))
	error_exit (_("Error reading events: %s"), error->message);
      g_clear_error (&error);
    }

  ceelog_source_free (src);

  ceelog_compiled_expr_free (conf_expr);

  return EXIT_SUCCESS;
}
